#include <libmbtk/config.h>
#include <stdlib.h>
#include <string.h>
#include "libmbtk.h"
#include "backend.h"
#include "display.h"
#include "event.h"
#include "widget.h"
#include "handler.h"

static const char *evname[] = {
	"KEY_PRESS",
	"KEY_RELEASE",
	"MOUSE_PRESS",
	"MOUSE_RELEASE",
	"MOUSE_MOTION",
	"MOUSE_SCROLL",
	"WIN_EXPOSE",
	"WIN_MOVE",
	"WIN_RESIZE",
	"WIN_CLOSE",
	"WIN_ENTER",
	"WIN_LEAVE",
	"STDIN",
	"FDWATCH",
	NULL
};

const char *mbtk_event_type_name(mbtk_event_type_t t)
{
	if ((t < 0) || (t >= MBTK_EV_max))
		return "unknown";
	return evname[(int)t];
}

int mbtk_be_iterate_(mbtk_display_t *disp, mbtk_event_t *dst, double timeout, int block)
{
	if ((disp == NULL) || (disp->be == NULL) || (disp->be->iterate == NULL))
		return -1;

	return disp->be->iterate(disp, dst, timeout, block);
}

void mbtk_watch_stdin_read(mbtk_display_t *disp, int enable)
{
	if ((disp == NULL) || (disp->be == NULL) || (disp->be->watch_stdin_read == NULL))
		return;

	return disp->be->watch_stdin_read(disp, enable);
}


void mbtk_watch_fd(mbtk_display_t *disp, int fd, int for_read, int for_write)
{
	if ((disp == NULL) || (disp->be == NULL) || (disp->be->watch_fd == NULL))
		return;

	return disp->be->watch_fd(disp, fd, for_read, for_write);
}

void mbtk_event_init_display(mbtk_display_t *disp)
{
	memset(&disp->evi, 0, sizeof(disp->evi));
	disp->evi.last_win_expose = disp->evi.last_win_resize = -100;
}

/* Return w or its first parent in the hierarchy that has handler for events.
   Returns NULL if no event handler is available in this subtree */
MBTK_INLINE mbtk_widget_t *first_handler_up(mbtk_widget_t *w)
{
	while(w != NULL) {
		if ((w->handler != NULL) && (w->handler->event != NULL))
			break;
		w = (mbtk_widget_t *)w->hvbox.parent;
	}
	return w;
}

MBTK_INLINE void call_handler(mbtk_widget_t *w, mbtk_event_t *ev)
{
	w = first_handler_up(w);
	if (w != NULL)
		w->handler->event(w, ev);
}

/* Return the widget for win;x;y (limit it to the first one in the
   hierarchy that has a handler if require_handler is 1) */
MBTK_INLINE mbtk_widget_t *widget_by_coord(mbtk_window_t *win, long x, long y, int require_handler)
{
	mbtk_widget_t *w = (mbtk_widget_t *)hvbox_by_coord(win->root, x, y);
	if (require_handler)
		w = first_handler_up(w);
	return w;
}

/*** High level ***/
#define EV_THROTTLE 0.03
#define WIN_REDRAW_TIMEOUT 0.03

int mbtk_iterate(mbtk_display_t *disp, mbtk_event_t *ev, int block)
{
	int res, blk;
	double now, dly, d;


	do {
		blk = block;
		dly = 10;
		now = mbtk_get_time_el(disp);
		if (disp->evi.exp_win_resize > 0) {
			d = disp->evi.exp_win_resize - now;
			if (d < dly) {
				dly = d;
				blk = 0;
			}
		}
		if (disp->evi.exp_win_expose > 0) {
			d = disp->evi.exp_win_expose - now;
			if (d < dly) {
				dly = d;
				blk = 0;
			}
		}
		if (disp->need_win_redraw) {
			if (disp->win_next_redraw == 0)
				disp->win_next_redraw = now + WIN_REDRAW_TIMEOUT;

			d = disp->win_next_redraw - now;
			if (d < dly) {
				dly = d;
				blk = 0;
			}
		}
		if (dly < 0)
			dly = 0;

		res = mbtk_be_iterate_(disp, ev, dly, blk);
		now = mbtk_get_time_el(disp);

		if ((disp->evi.exp_win_resize > 0) && (now >= disp->evi.exp_win_resize)) {
			mbtk_win_resized(disp->evi.ev_win_resize.win, disp->evi.ev_win_resize.data.win.w, disp->evi.ev_win_resize.data.win.h);
			disp->evi.last_win_resize = now;
			disp->evi.exp_win_resize = 0;
		}

		if ((disp->evi.exp_win_expose > 0) && (now >= disp->evi.exp_win_expose)) {
			mbtk_win_redraw(disp->evi.ev_win_expose.win);
			disp->evi.last_win_expose = now;
			disp->evi.exp_win_expose = 0;
		}

		if ((disp->need_win_redraw) && (now >= disp->win_next_redraw)) {
			mbtk_display_redraw_windows(disp);
			disp->win_next_redraw = 0;
			disp->need_win_redraw = 0;
		}

		if (res > 0) {
			block = 0;
			switch(ev->type) {
				case MBTK_EV_WIN_MOVE:
					mbtk_win_moved(ev->win, ev->data.win.x, ev->data.win.y);
					break;
				case MBTK_EV_WIN_EXPOSE:
					ev->win->draw_dirty = 1;
					if ((disp->evi.exp_win_expose > 0) && (disp->evi.ev_win_expose.win != ev->win)) {
						/* if an already running timer uses another window, flush the old first */
						mbtk_win_redraw(disp->evi.ev_win_expose.win);
					}
					if (disp->evi.last_win_expose >= (now - EV_THROTTLE)) {
						disp->evi.ev_win_expose = *ev;
						if (disp->evi.exp_win_expose <= 0)
							disp->evi.exp_win_expose = now + EV_THROTTLE;
					}
					else {
						/* first event after a while */
						mbtk_win_redraw(ev->win);
						disp->evi.last_win_expose = now;
					}
					break;
				case MBTK_EV_WIN_RESIZE:
					if ((disp->evi.exp_win_resize > 0) && (disp->evi.ev_win_resize.win != ev->win)) {
						/* if an already running timer uses another window, flush the old first */
						mbtk_win_resized(disp->evi.ev_win_resize.win, disp->evi.ev_win_resize.data.win.w, disp->evi.ev_win_resize.data.win.h);
					}
					if (disp->evi.last_win_resize >= (now - EV_THROTTLE)) {
						disp->evi.ev_win_resize = *ev;
						if (disp->evi.exp_win_resize <= 0)
							disp->evi.exp_win_resize = now + EV_THROTTLE;
					}
					else {
						/* first event after a while */
						mbtk_win_resized(ev->win, ev->data.win.w, ev->data.win.h);
						disp->evi.last_win_resize = now;
					}
					break;

					case MBTK_EV_WIN_LEAVE:
						if (ev->win->focus_has_pointer) {
							ev->win->focus_has_pointer = 0;
							ev->type = MBTK_EV_FOCUS_LEAVE;
							call_handler(ev->win->focus, ev);
							ev->type = MBTK_EV_WIN_LEAVE;
						}
						break;

					case MBTK_EV_MOUSE_MOTION:
						if (ev->win != NULL) {
							mbtk_widget_t *w = widget_by_coord(ev->win, ev->data.mouse.x, ev->data.mouse.y, 1);
							call_handler(w, ev);
							if ((ev->win != NULL) && (ev->win->focus != NULL)) {
								if ((w == ev->win->focus) && !ev->win->focus_has_pointer) {
									ev->win->focus_has_pointer = 1;
									ev->type = MBTK_EV_FOCUS_ENTER;
									call_handler(ev->win->focus, ev);
									ev->type = MBTK_EV_MOUSE_MOTION;
								}
								else if ((w != ev->win->focus) && ev->win->focus_has_pointer) {
									ev->win->focus_has_pointer = 0;
									ev->type = MBTK_EV_FOCUS_LEAVE;
									call_handler(ev->win->focus, ev);
									ev->type = MBTK_EV_MOUSE_MOTION;
								}
							}
						}
						break;

					case MBTK_EV_MOUSE_PRESS:
					case MBTK_EV_MOUSE_RELEASE:
					case MBTK_EV_MOUSE_SCROLL:
						if (ev->win != NULL) {
							mbtk_widget_t *w = widget_by_coord(ev->win, ev->data.mouse.x, ev->data.mouse.y, 1);
							call_handler(w, ev);
							if (ev->type == MBTK_EV_MOUSE_RELEASE)
								ev->win->origin = NULL;
						}
						break;
					case MBTK_EV_KEY_PRESS:
					case MBTK_EV_KEY_RELEASE:
						call_handler(ev->win->focus, ev);
						break;
				default:
					break;
			}
		}

	} while(block);
	return res;
}

