#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "hvbox.h"

hvbox_t *hvbox_new_root(int w, int h, void *widget)
{
	hvbox_t *r;
	r = calloc(sizeof(hvbox_t), 1);
	r->w_min = r->w = w;
	r->h_min = r->h = h;
	r->type = HVBOX_ROOT;
	return r;
}

/* Allocate a new node and fill in w_min and h_min */
#define hvbox_new(b, wmin, hmin) \
do { \
	b = calloc(sizeof(hvbox_t), 1); \
	if (wmin != 0) \
		b->w_min = wmin; \
	if (hmin != 0) \
		b->h_min = hmin; \
} while(0)

/* append a node under a parent as the last child */
static hvbox_t *hvbox_append(hvbox_t *parent, hvbox_t *b)
{
	if (parent->ch_first == NULL)
		parent->ch_first = b;
	else
		parent->ch_last->next = b;
	parent->ch_last = b;
	b->parent = parent;
	return b;
}

/* insert a node under a parent as the first child */
static hvbox_t *hvbox_insert(hvbox_t *parent, hvbox_t *b)
{
	if (parent->ch_last == NULL)
		parent->ch_last = b;

	b->next = parent->ch_first;
	parent->ch_first = b;
	b->parent = parent;
	return b;
}

/* append a node after a sibling */
static hvbox_t *hvbox_append_at(hvbox_t *sibling, hvbox_t *b)
{
	if (sibling->parent->ch_last == sibling)
		sibling->parent->ch_last = b;

	b->parent = sibling->parent;
	b->next = sibling->next;
	sibling->next = b;
	return b;
}

/* search for the previous box before b in the list of siblings */
static hvbox_t *hvbox_prev(hvbox_t *b)
{
	hvbox_t *prev, *n;
	for(prev = NULL, n = b->parent->ch_first; (n != b) || (n == NULL); prev = n, n = n->next) ;
	return prev;
}

/* insert a node before a sibling */
static hvbox_t *hvbox_insert_at(hvbox_t *sibling, hvbox_t *b)
{
	hvbox_t *prev = hvbox_prev(sibling);

	if (prev != NULL)
		prev->next = b;

	if (sibling->parent->ch_first == sibling)
		sibling->parent->ch_first = b;

	b->parent = sibling->parent;
	b->next = sibling;
	return b;
}

hvbox_t *hvbox_append_widget(hvbox_t *parent, int w_min, int h_min, void *widget)
{
	hvbox_t *b;
	hvbox_new(b, w_min, h_min);
	b->widget = widget;
	return hvbox_append(parent, b);
}

hvbox_t *hvbox_insert_widget(hvbox_t *parent, int w_min, int h_min, void *widget)
{
	hvbox_t *b;
	hvbox_new(b, w_min, h_min);
	b->widget = widget;
	return hvbox_insert(parent, b);
}

hvbox_t *hvbox_append_box(hvbox_t *parent, hvbox_type_t type)
{
	hvbox_t *b;
	hvbox_new(b, 0, 0);
	b->type = type;
	return hvbox_append(parent, b);
}

hvbox_t *hvbox_insert_box(hvbox_t *parent, hvbox_type_t type)
{
	hvbox_t *b;
	hvbox_new(b, 0, 0);
	b->type = type;
	return hvbox_insert(parent, b);
}

hvbox_t *hvbox_append_existing_box(hvbox_t *parent, hvbox_t *box)
{
	if (box->parent != NULL)
		return NULL;
	return hvbox_append(parent, box);
}

hvbox_t *hvbox_insert_existing_box(hvbox_t *parent, hvbox_t *box)
{
	if (box->parent != NULL)
		return NULL;
	return hvbox_insert(parent, box);
}

hvbox_t *hvbox_append_existing_box_at(hvbox_t *sibling, hvbox_t *box)
{
	if (box->parent != NULL)
		return NULL;
	return hvbox_append_at(sibling, box);
}

hvbox_t *hvbox_insert_existing_box_at(hvbox_t *sibling, hvbox_t *box)
{
	if (box->parent != NULL)
		return NULL;
	return hvbox_insert_at(sibling, box);
}



hvbox_t *hvbox_new_floating_box(int w_min, int h_min, hvbox_type_t type)
{
	hvbox_t *b;
	hvbox_new(b, w_min, h_min);
	b->type = type;
	return b;
}

static void hvbox_free(hvbox_t *from)
{
	hvbox_t *n;
	for(n = from->ch_first; n != NULL; n = n->next)
		hvbox_free(n);
	free(from);
}

void hvbox_unlink(hvbox_t *from)
{
	/* unlink (if from is not the root) */
	if (from->parent != NULL) {
		hvbox_t *prev, *n;
		for(prev = NULL, n = from->parent->ch_first; (n != from) || (n == NULL); prev = n, n = n->next) ;
		assert(n == from); /* else parent doesn't even have our node */
		if (prev != NULL)
			prev->next = n->next;
		if (from->parent->ch_first == n)
			from->parent->ch_first = n->next;
		if (from->parent->ch_last == n)
			from->parent->ch_last = prev;
		from->parent = NULL;
	}
}

void hvbox_destroy(hvbox_t *from)
{
	hvbox_unlink(from);

	/* recursive free */
	hvbox_free(from);
}

/* propagate minimal sizes from widget leaves upwards */
static void hvbox_update_prop_min_up(hvbox_t *node)
{
	hvbox_t *n;

	node->ox = node->x;
	node->oy = node->y;
	node->ow = node->w;
	node->ow = node->h;

	if (node->hidden) {
		node->w_min_calc = node->w = node->h_min_calc = node->h = 0;
		return;
	}

	for(n = node->ch_first; n != NULL; n = n->next)
		hvbox_update_prop_min_up(n);
	
	switch(node->type) {
		case HVBOX_HBOX:
			node->w = node->fr_l + node->fr_r; node->h = node->fr_t + node->fr_b;
			for(n = node->ch_first; n != NULL; n = n->next) {
				node->w += n->w;
				if (n->h > node->h)
					node->h = n->h;
				if (n->next != NULL)
					node->w += node->spacing;
			}
			break;
		case HVBOX_VBOX:
			node->w = node->fr_l + node->fr_r; node->h = node->fr_t + node->fr_b;
			for(n = node->ch_first; n != NULL; n = n->next) {
				node->h += n->h;
				if (n->w > node->w)
					node->w = n->w;
				if (n->next != NULL)
					node->w += node->spacing;
			}
			break;
		case HVBOX_WIDGET:
			node->w_min_calc = node->w = node->w_min + 2 * node->padding + node->fr_l + node->fr_r;
			node->h_min_calc = node->h = node->h_min + 2 * node->padding + node->fr_t + node->fr_b;
			return;
		case HVBOX_ROOT:
			if (node->ch_first != NULL) {
				node->w_min_calc = node->ch_first->w_min_calc + 2*node->padding + node->fr_l + node->fr_r;
				node->h_min_calc = node->ch_first->h_min_calc + 2*node->padding + node->fr_t + node->fr_b;
				return;
			}
			break;
	}
	node->w += 2*node->padding;
	node->h += 2*node->padding;
	if (node->w < node->w_min)
		node->w = node->w_min;
	if (node->h < node->h_min)
		node->h = node->h_min;
	node->w_min_calc = node->w;
	node->h_min_calc = node->h;
}

static void hvbox_update_expand(hvbox_t *node, hvbox_span_t alt_span, int *alt_size, int parent_size)
{
	if (node->hidden) {
		*alt_size = 0;
	}
	else if (alt_span == HVBOX_FILL) {
		if (*alt_size < parent_size)
			*alt_size = parent_size;
	}
}

/* propagate calculated sizes downwards */
static void hvbox_update_prop_calc_down(hvbox_t *node, int force_hidden)
{
	hvbox_t *n;

	if ((!node->hidden) && (!force_hidden)) {
		switch(node->type) {
			case HVBOX_WIDGET:
				node->hidden_calc = force_hidden;
				return;
			case HVBOX_HBOX:
				for(n = node->ch_first; n != NULL; n = n->next)
					n->h = node->h - 2*node->padding - node->fr_t - node->fr_b;
				break;
			case HVBOX_VBOX:
				for(n = node->ch_first; n != NULL; n = n->next)
					n->w = node->w - 2*node->padding - node->fr_l - node->fr_r;
				hvbox_update_expand(node, node->alt_span, &node->h, node->parent->h);
				break;
			case HVBOX_ROOT:
				for(n = node->ch_first; n != NULL; n = n->next) {
						n->h = node->h - node->fr_l - node->fr_r;
						n->w = node->w - node->fr_t - node->fr_b;
				}
				break;
		}
	}
	else
		force_hidden = 1;

	node->hidden_calc = force_hidden;

	node->num_ch = 0;
	for(n = node->ch_first; n != NULL; n = n->next) {
		hvbox_update_prop_calc_down(n, force_hidden);
		node->num_ch++;
	}
}


/* Calculate size of % widgets and count fillers and accumulated total size */
static void hvbox_update_fill1(hvbox_t *node, int *node_size, hvbox_span_t span, int parent_size, int *fills, int *used)
{
	if (node->hidden) {
		*node_size = 0;
		return;
	}
	if ((span > 0) && (span <= 100))
		*node_size = parent_size * span / 100;
	else if (span == HVBOX_FILL)
		(*fills)++;
	(*used) += *node_size;
}

/* Enlarge fillers to evenly take up remaining space */
static void hvbox_update_fill2(hvbox_t *node, int *node_size, hvbox_span_t span, int parent_size, int fills, int used, int *avail, int extra)
{
	if (node->hidden)
		return;
	switch(span) {
		case HVBOX_MINIMAL:
			break;
		case HVBOX_FILL:
			if (extra > *avail)
				extra = *avail;
			*node_size += extra;
			*avail -= extra;
			break;
		default:
			/* percent, already done */
			break;
	}
}

/* determine who fills the the remaining space */
static void hvbox_update_fill(hvbox_t *node)
{
	hvbox_t *n;
	int fills = 0, used = 0, extra = 0, spc = 0;

	if ((node->num_ch > 1) && (node->spacing > 0))
		spc = (node->num_ch - 1) * node->spacing;

	switch(node->type) {
		case HVBOX_HBOX:
			for(n = node->ch_first; n != NULL; n = n->next)
				hvbox_update_fill1(n, &n->w, n->span, node->w, &fills, &used);
			hvbox_update_expand(node, node->alt_span, &node->w, node->parent->w);
			node->avail = node->w - used - 2*node->padding - node->fr_l - node->fr_r - spc;
			if (fills > 0)
				extra = node->avail / fills;
			for(n = node->ch_first; n != NULL; n = n->next)
				hvbox_update_fill2(n, &n->w, n->span, node->w, fills, used, &node->avail, extra);
			if (fills > 0)
				node->avail = 0;
			break;
		case HVBOX_VBOX:
			for(n = node->ch_first; n != NULL; n = n->next)
				hvbox_update_fill1(n, &n->h, n->span, node->h, &fills, &used);
			hvbox_update_expand(node, node->alt_span, &node->h, node->parent->h);
			node->avail = node->h - used - 2*node->padding - node->fr_t - node->fr_b - spc;
			if (fills > 0)
				extra = node->avail / fills;
			for(n = node->ch_first; n != NULL; n = n->next)
				hvbox_update_fill2(n, &n->h, n->span, node->h, fills, used, &node->avail, extra);
			if (fills > 0)
				node->avail = 0;
			break;
		case HVBOX_WIDGET:
			node->avail = 0;
			break;
		case HVBOX_ROOT:
			break;
	}

	for(n = node->ch_first; n != NULL; n = n->next)
		hvbox_update_fill(n);
}

/* determine the coordinates */
static int hvbox_update_coords(hvbox_t *node)
{
	hvbox_t *n;
	int c, c_spc;

	if (node->hidden_calc)
		return 0;

	switch(node->type) {
		case HVBOX_HBOX: c = node->x + node->padding + node->fr_l; break;
		case HVBOX_VBOX: c = node->y + node->padding + node->fr_t; break;
		case HVBOX_ROOT:
		case HVBOX_WIDGET: c = 0; break;
	}

	c_spc = 0;
	if (node->avail > 0) {
		switch(node->align) {
			case HVBOX_LEFT_TOP:
				break;
			case HVBOX_RIGHT_BOTTOM:
				c += node->avail;
				break;
			case HVBOX_JUSTIFY:
				if (node->num_ch > 1) {
					c_spc = node->avail/(node->num_ch-1);
					break;
				}
				/* else fall through and use center */
			case HVBOX_CENTER:
				c += node->avail/2;
				break;
/*			case HVBOX_JUSTIFY: ... with space around the row
				c_spc = node->avail/(node->num_ch+1);
				c += c_spc;
				break; */
		}
	}

	switch(node->type) {
		case HVBOX_HBOX:
			for(n = node->ch_first; n != NULL; n = n->next) {
				n->y = node->y + node->padding + node->fr_t;
				n->x = c;
				c += n->w + c_spc + node->spacing;
			}
			break;
		case HVBOX_VBOX:
			for(n = node->ch_first; n != NULL; n = n->next) {
				n->x = node->x + node->padding + node->fr_l;
				n->y = c;
				c += n->h + c_spc + node->spacing;
			}
			break;
		case HVBOX_WIDGET:
			/* widget's own padding should offset it's top-left corner */
			node->x += node->padding + node->fr_l;
			node->y += node->padding + node->fr_t;
			node->h -= 2*node->padding - node->fr_l - node->fr_r;
			node->w -= 2*node->padding - node->fr_t - node->fr_b;
			break;
		case HVBOX_ROOT:
			if (node->ch_first != NULL) {
				node->ch_first->x = node->padding + node->fr_l;
				node->ch_first->y = node->padding + node->fr_t;
			}
			break;
	}

	for(n = node->ch_first; n != NULL; n = n->next)
		hvbox_update_coords(n);

	if (!node->dirty) {
		if ((node->ox != node->x) || (node->oy != node->y) || (node->ow != node->w) || (node->ow != node->h))
			node->dirty = 1;
	}

	return 0;
}

int hvbox_update(hvbox_t *root)
{
	hvbox_update_prop_min_up(root);
	hvbox_update_prop_calc_down(root, 0);
	hvbox_update_fill(root);
	return hvbox_update_coords(root);
}

#define within(n, cx, cy) (((cx) >= (n)->x) && ((cy) >= (n)->y) && ((cx) <= ((n)->x + (n)->w)) && ((cy) <= ((n)->y + (n)->h)))

hvbox_t *hvbox_by_coord(hvbox_t *root, int x, int y)
{
	hvbox_t *found;
	int descend;
	if (!within(root, x, y))
		return NULL;
	found = root;
	do {
		hvbox_t *n;
		descend = 0;
		for(n = found->ch_first; n != NULL; n = n->next) {
			if (within(n, x, y)) {
				found = n;
				descend = 1;
				break;
			}
		}
	} while(descend);
	return found;
}


/* add a node to the open list of a specfici level */
static void iter_set(hvbox_iter_t *iter, int level, hvbox_t *node)
{
	int idx = level % 2;
	if (iter->used[idx] >= iter->alloced[idx]) {
		iter->alloced[idx] += 32;
		iter->open[idx] = realloc(iter->open[idx], sizeof(hvbox_t) * iter->alloced[idx]);
	}
	(iter->open[idx])[iter->used[idx]] = node;
	iter->used[idx]++;
}

/* get (pop) a node from the open list of a specific level */
static hvbox_t *iter_get(hvbox_iter_t *iter, int level)
{
	int idx = level % 2;
	if (iter->cursor[idx] < iter->used[idx]) {
		hvbox_t *r;
		r = (iter->open[idx])[iter->cursor[idx]];
		iter->cursor[idx]++;
		return r;
	}
	return NULL;
}

/* reset a level (make it empty) */
static void iter_reset(hvbox_iter_t *iter, int level)
{
	iter->used[level % 2] = 0;
	iter->cursor[level % 2] = 0;
}

/* free the iterator */
static void iter_destroy(hvbox_iter_t *iter)
{
	if (iter->open[0] != 0)
		free(iter->open[0]);
	if (iter->open[1] != 0)
		free(iter->open[1]);
}

hvbox_t *hvbox_bfs_first(hvbox_iter_t *iterator, hvbox_t *node)
{
	memset(iterator, 0, sizeof(hvbox_iter_t));

	if (node != NULL) {
		iter_set(iterator, 0, node);
		return hvbox_bfs_next(iterator);
	}
	return NULL;
}

hvbox_t *hvbox_bfs_next(hvbox_iter_t *iterator)
{
	hvbox_t *r, *n;

	/* get next node from the open list of the current level */
	r = iter_get(iterator, iterator->level);

	if (r == NULL) {
		/* no more - reset current level and switch to the next level */
		iter_reset(iterator, iterator->level);
		iterator->level++;
		r = iter_get(iterator, iterator->level);
		if (r == NULL)
			return NULL;
	}

	/* add each children of the current node to the next open level */
	for(n = r->ch_first; n != NULL; n = n->next)
		iter_set(iterator, iterator->level+1, n);
	return r;
}

void hvbox_bfs_finish(hvbox_iter_t *iterator)
{
	iter_destroy(iterator);
}

